import re
from typing import Any, Callable, TypeVar, cast

T = TypeVar("T", bound=dict[str, Any])

camel_case_pattern = re.compile(r"(?<!^)(?=[A-Z])")


def camel_to_snake_case(text: str) -> str:
    return camel_case_pattern.sub("_", text).lower()


def change_dict_keys(data: T, func: Callable[[str], str]) -> T:
    new_data: T = cast(T, {})
    for key, value in data.items():
        if isinstance(value, dict):
            new_data[func(key)] = change_dict_keys(value, func)
        elif isinstance(value, list):
            new_array = []
            for v in value:
                if isinstance(v, dict):
                    new_array.append(change_dict_keys(v, func))
                else:
                    new_array.append(v)
            new_data[func(key)] = new_array
        else:
            new_data[func(key)] = value
    return new_data



import yaml

from ..iparser import IParser, PARSER_RETURN_TYPE


class YamlParser(IParser):
    def parse(self, value: str) -> PARSER_RETURN_TYPE:
        return yaml.safe_load(value)  # type: ignore [no-any-return]



from __future__ import annotations


class TemplateCompiler:
    values: dict[str, str]

    def __init__(self) -> None:
        self.values = {}

    @staticmethod
    def build_template_name(name: str) -> str:
        return f"<{name}>"

    def set(self, name: str, value: str) -> TemplateCompiler:
        self.values[name] = value
        return self

    def set_many(self, values: dict[str, str]) -> TemplateCompiler:
        self.values = {**self.values, **values}
        return self

    def compile(self, template: str) -> str:
        for name, value in self.values.items():
            template = template.replace(self.build_template_name(name), value)
        return template



from ..iloader import ILoader


class FileLoader(ILoader):
    def __init__(self, path: str) -> None:
        self.path = path

    def load(self) -> str:
        with open(self.path, "r") as file:
            return file.read()



from abc import ABC, abstractmethod
from typing import Any

PARSER_RETURN_TYPE = dict[str, Any]


class IParser(ABC):
    @abstractmethod
    def parse(self, value: str) -> PARSER_RETURN_TYPE:
        pass



from string import Template
from typing import Any


SCRIPT_TEMPLATE = Template(
    """
    while true; do if $ready_script; then break;
    else echo "Ready script exited with 1, wait 3 seconds";
    sleep 3; fi; done; $main_command
"""
)


class JobAdapter:
    def adapt(self, job: dict[str, Any]) -> dict[str, Any]:
        annotations = job["metadata"].get("annotations", {})
        if "ktest/ready-script" not in annotations:
            raise ValueError("ktest/ready-script annotation is not provided")

        for container in job["spec"]["template"]["spec"]["containers"]:
            if container["name"] == "main":
                main_command = container["command"] + container["args"]
                container["command"] = ["sh", "-c"]
                container["args"] = [
                    SCRIPT_TEMPLATE.substitute(
                        ready_script=annotations["ktest/ready-script"],
                        main_command=" ".join(main_command),
                    )
                ]
                continue
            shutdown_script_key = f"ktest/shutdown-{container['name']}"
            if shutdown_script_key not in annotations:
                raise ValueError(
                    f"{shutdown_script_key} annotation is not provided"
                )

        return job



import time

from kubernetes import client
from kubernetes.stream import stream


class KTestClient:
    def __init__(
        self,
        k_client: client.ApiClient,
        namespace: str,
        poll_timeout_s: int,
    ) -> None:
        self.k_client = k_client
        self.namespace = namespace
        self.poll_timeout_s = poll_timeout_s

    def run_test_job(self, body: client.V1Job) -> int:
        batch_api = client.BatchV1Api(self.k_client)
        core_api = client.CoreV1Api(self.k_client)
        job = batch_api.create_namespaced_job(
            body=body, namespace=self.namespace, pretty=True
        )
        print(f"Job {job.metadata.name} is created")
        pod = self.get_job_pod(core_api, job.metadata.name)
        print(f"Job pod {pod.metadata.name} is created")
        self.wait_for_main_to_start(core_api, pod.metadata.name)
        print("Main container started, start streaming logs...")
        for line in core_api.read_namespaced_pod_log(
            name=pod.metadata.name,
            namespace=self.namespace,
            container="main",
            follow=True,
            _preload_content=False,
        ).stream():
            print(line.decode("utf-8"), end="")

        exit_code = self.get_main_container_exit_code(
            core_api, pod.metadata.name
        )
        print(f"Main container exited with status code: {exit_code}")
        for container in pod.spec.containers:
            if container.name == "main":
                continue
            shutdown_script = job.metadata.annotations[
                f"ktest/shutdown-{container.name}"
            ]
            print(
                f"Shutting down {container.name} container "
                f"with script: {shutdown_script}"
            )
            resp = stream(
                core_api.connect_post_namespaced_pod_exec,
                name=pod.metadata.name,
                namespace=self.namespace,
                container=container.name,
                command=["sh", "-c", shutdown_script],
                stderr=True,
                stdout=True,
                stdin=False,
                tty=False,
            )
            print(f"{container.name}: ", resp)
        return exit_code

    def wait_for_main_to_start(
        self, core_api: client.CoreV1Api, pod_name: str
    ) -> None:
        while True:
            pod = core_api.read_namespaced_pod_status(
                name=pod_name, namespace=self.namespace
            )
            for c in pod.status.container_statuses:
                if c.name == "main" and c.started:
                    return
            print("Wait for main container to start...")
            time.sleep(self.poll_timeout_s)

    def get_main_container_exit_code(
        self,
        core_api: client.CoreV1Api,
        pod_name: str,
    ) -> int:
        while True:
            pod = core_api.read_namespaced_pod_status(
                name=pod_name, namespace=self.namespace
            )
            for c in pod.status.container_statuses:
                if c.name == "main" and c.state.terminated is not None:
                    return int(c.state.terminated.exit_code)
            print("Wait for main container to stop...")
            time.sleep(self.poll_timeout_s)

    def get_job_pod(
        self, core_api: client.CoreV1Api, job_name: str
    ) -> client.V1Pod:
        while True:
            response = core_api.list_namespaced_pod(
                label_selector=f"job-name={job_name}",
                namespace=self.namespace,
            )
            if len(response.items) > 0:
                return response.items[0]
            print("Waiting for job pod to create...")
            time.sleep(self.poll_timeout_s)



import os
from pathlib import Path

from setuptools import setup  # type: ignore [import]


setup(
    name="kube-tests",
    version=os.environ["GITHUB_REF_NAME"],
    description="Kubernetes test client",
    author="Vladimir Vojtenko",
    author_email="vladimirdev635@gmail.com",
    license="MIT",
    packages=["kube_tests", "kube_tests.loaders", "kube_tests.parsers"],
    package_data={
        "kube_tests": ["py.typed"],
        "kube_tests.loaders": ["py.typed"],
        "kube_tests.parsers": ["py.typed"],
    },
    install_requires=["kubernetes", "pyyaml"],
    long_description=(Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
)



from .adapter import JobAdapter
from .client import KTestClient
from .iloader import ILoader
from .iparser import IParser
from .job_builder import JobBuilder
from .loaders import FileLoader
from .parsers import YamlParser
from .template_compiler import TemplateCompiler


__all__ = (
    "KTestClient",
    "FileLoader",
    "YamlParser",
    "JobAdapter",
    "TemplateCompiler",
    "JobBuilder",
    "ILoader",
    "IParser",
)



from __future__ import annotations

from typing import Any

from .adapter import JobAdapter
from .iloader import ILoader
from .iparser import IParser
from .template_compiler import TemplateCompiler


class JobBuilder:
    job: dict[str, Any]
    _adapter: JobAdapter | None
    _loader: ILoader | None
    _parser: IParser | None
    _template_compiler: TemplateCompiler | None

    def __init__(self) -> None:
        self.job = {}
        self._loader = None
        self._parser = None
        self._template_compiler = None
        self._adapter = None

    def loader(self, loader: ILoader) -> JobBuilder:
        self._loader = loader
        return self

    def parser(self, parser: IParser) -> JobBuilder:
        self._parser = parser
        return self

    def template_compiler(self, compiler: TemplateCompiler) -> JobBuilder:
        self._template_compiler = compiler
        return self

    def adapter(self, adapter: JobAdapter) -> JobBuilder:
        self._adapter = adapter
        return self

    def build(self) -> dict[str, Any]:
        assert self._loader is not None, "Loader is not provided"
        assert (
            self._template_compiler is not None
        ), "Template compiler is not provided"
        assert self._parser is not None, "Parser is not provided"
        assert self._adapter is not None, "Adapter is not provided"
        source_template = self._loader.load()
        compiled_template = self._template_compiler.compile(source_template)
        job = self._parser.parse(compiled_template)
        return self._adapter.adapt(job)
